
if not GroupAIStateBesiege then
	return
end

function GroupAIStateBesiege:_is_assault_active()
	return self._task_data.assault.active
end
function GroupAIStateBesiege:_init_misc_data()
	GroupAIStateBesiege.super:_init_misc_data(self)
	if managers.navigation:is_data_ready() then
		self._nr_dynamic_waves = 0
		self._nr_waves = 0
		self:_create_area_data()
		self._task_data = {}
		self._spawn_cap_pressure = 13
		self._fed_times = 1
		self._dead_civs = 0
		self._rate = 0
		self._dead_cameras = 0
		self._dead_civs_unlimited = 0
		self._needing_for_special_units = 0
		self._grenade_pressure_multiplier = 1
		self._task_data.reenforce = {
			tasks = {},
			next_dispatch_t = 0
		}
		self._task_data.recon = {
			tasks = {},
			next_dispatch_t = 0
		}
		self._task_data.assault = {disabled = true, is_first = true}
		self._task_data.regroup = {}
		local all_areas = self._area_data
		for u_key, u_data in pairs(self._police) do
			if not u_data.assigned_area then
				local nav_seg = u_data.unit:movement():nav_tracker():nav_segment()
				self:set_enemy_assigned(self:get_area_from_nav_seg_id(nav_seg), u_key)
			end
		end
	end
end

function GroupAIStateBesiege:_increase_fed_time()
	self._fed_times = self._fed_times + 1
end

function GroupAIStateBase:_change_grenade_pressure(amount)
	self._grenade_pressure_multiplier = self._grenade_pressure_multiplier + amount
end

function GroupAIStateBesiege:_check_fed_time()
	local needed_times = math.lerp(60, 100, math.random())
	if self._fed_times >= needed_times then
		self._fed_times = 1
		return true
	else
		return false
	end
end

function UnitPos()
	local pos
	for criminal_key, criminal_data in pairs( managers.groupai:state():all_player_criminals() ) do
		if not criminal_data.status then
			pos = criminal_data.m_pos
		end
	end
	return pos
end

function Unit_itself()
	local pos
	for criminal_key, criminal_data in pairs( managers.groupai:state():all_player_criminals() ) do
		if not criminal_data.status then
			the_data = criminal_data
		end
	end
	return the_data
end

function GroupAIStateBesiege:_set_assault_objective_to_group(group, phase)
	if not group.has_spawned then
		return
	end
	local phase_is_anticipation = phase == "anticipation"
	local current_objective = group.objective
	local approach, open_fire, push, pull_back, charge
	local obstructed_area = self:_chk_group_areas_tresspassed(group)
	local group_leader_u_key, group_leader_u_data = self._determine_group_leader(group.units)
	local tactics_map
	if group_leader_u_data and group_leader_u_data.tactics then
		tactics_map = {}
		for _, tactic_name in ipairs(group_leader_u_data.tactics) do
			tactics_map[tactic_name] = true
		end
		if current_objective.tactic and not tactics_map[current_objective.tactic] then
			current_objective.tactic = nil
		end
		for i_tactic, tactic_name in ipairs(group_leader_u_data.tactics) do
			if tactic_name == "deathguard" and not phase_is_anticipation then
				if current_objective.tactic == tactic_name then
					for u_key, u_data in pairs(self._char_criminals) do
						if u_data.status and current_objective.follow_unit == u_data.unit then
							local crim_nav_seg = u_data.tracker:nav_segment()
							if current_objective.area.nav_segs[crim_nav_seg] then
								return
							end
						end
					end
				end
				local closest_crim_u_data, closest_crim_dis_sq
				for u_key, u_data in pairs(self._char_criminals) do
					if u_data.status then
						local closest_u_id, closest_u_data, closest_u_dis_sq = self._get_closest_group_unit_to_pos(u_data.m_pos, group.units)
						if closest_u_dis_sq and (not closest_crim_dis_sq or closest_crim_dis_sq > closest_u_dis_sq) then
							closest_crim_u_data = u_data
							closest_crim_dis_sq = closest_u_dis_sq
						end
					end
				end
				if closest_crim_u_data then
					local search_params = {
						from_tracker = group_leader_u_data.unit:movement():nav_tracker(),
						to_tracker = closest_crim_u_data.tracker,
						id = "GroupAI_deathguard",
						access_pos = self._get_group_acces_mask(group)
					}
					local coarse_path = managers.navigation:search_coarse(search_params)
					if coarse_path then
						local grp_objective = {
							type = "assault_area",
							tactic = "deathguard",
							follow_unit = closest_crim_u_data.unit,
							distance = 800,
							area = self:get_area_from_nav_seg_id(coarse_path[#coarse_path][1]),
							coarse_path = coarse_path,
							attitude = "engage",
							moving_in = true
						}
						group.is_chasing = true
						self:_set_objective_to_enemy_group(group, grp_objective)
						self:_voice_deathguard_start(group)
						return
					end
				end
			elseif tactic_name == "charge" and not current_objective.moving_out and group.in_place_t and (self._t - group.in_place_t > 15 or self._t - group.in_place_t > 4 and self._drama_data.amount <= tweak_data.drama.low) and next(current_objective.area.criminal.units) and group.is_chasing and not current_objective.charge then
				charge = true
			end
		end
	end
	local objective_area
	if obstructed_area then
		if current_objective.moving_out then
			if not current_objective.open_fire then
				open_fire = true
			end
		elseif not current_objective.pushed or charge and not current_objective.charge then
			push = true
		end
	else
		local obstructed_path_index = self:_chk_coarse_path_obstructed(group)
		if obstructed_path_index then
			print("obstructed_path_index", obstructed_path_index)
			objective_area = self:get_area_from_nav_seg_id(group.coarse_path[math.max(obstructed_path_index - 1, 1)][1])
			pull_back = true
		elseif not current_objective.moving_out then
			local has_criminals_close
			if not current_objective.moving_out then
				for area_id, neighbour_area in pairs(current_objective.area.neighbours) do
					if next(neighbour_area.criminal.units) then
						has_criminals_close = true
					else
					end
				end
			end
			if charge then
				push = true
			elseif not push and not phase_is_anticipation then
				approach = true
			elseif not phase_is_anticipation and not group.in_place_t and not current_objective.open_fire then
				open_fire = true
			elseif not phase_is_anticipation and group.in_place_t and (group.is_chasing or self._t - group.in_place_t > 4) or not phase_is_anticipation and has_criminals_close then
				push = true
			elseif phase_is_anticipation and current_objective.open_fire then
				pull_back = true
			end
		end
	end
	objective_area = objective_area or current_objective.area
	if open_fire then
		push = true
		self:_voice_open_fire_start(group)
	elseif approach or push then
		local criminal_nav_seg
		for criminal_key, criminal_data in pairs(self._player_criminals) do
			if not criminal_data.status then
				criminal_nav_seg = self:get_area_from_nav_seg_id(criminal_data.tracker:nav_segment())
			end
		end
		local assault_area, alternate_assault_area, alternate_assault_area_from, assault_path, alternate_assault_path
		local to_search_areas = {objective_area}
		local found_areas = {
			[objective_area] = "init"
		}
		repeat
			local search_area = table.remove(to_search_areas, 1)
			if next(search_area.criminal.units) then
				local assault_from_here = true
				if not push and tactics_map and tactics_map.flank then
					local assault_from_area = found_areas[search_area]
					if assault_from_area ~= "init" then
						local cop_units = assault_from_area.police.units
						for u_key, u_data in pairs(cop_units) do
							if u_data.group and u_data.group ~= group and u_data.group.objective.type == "assault_area" then
								assault_from_here = false
								if not alternate_assault_area or math.random() < 0.5 then
								
									local search_params = {
										from_seg = current_objective.area.pos_nav_seg,
										to_seg = search_area.pos_nav_seg,
										id = "GroupAI_assault",
										access_pos = self._get_group_acces_mask(group),
										verify_clbk = callback(self, self, "is_nav_seg_safe")
									}
									alternate_assault_path = managers.navigation:search_coarse(search_params)
									if alternate_assault_path then
										self:_merge_coarse_path_by_area(alternate_assault_path)
										alternate_assault_area = search_area
										alternate_assault_area_from = assault_from_area
									end
								end
								found_areas[search_area] = nil
							else
							end
						end
					end
				end
				if assault_from_here then
					
					local search_params = {
						from_seg = current_objective.area.pos_nav_seg,
						to_seg = search_area.pos_nav_seg,
						id = "GroupAI_assault",
						access_pos = self._get_group_acces_mask(group),
						verify_clbk = callback(self, self, "is_nav_seg_safe")
					}
					assault_path = managers.navigation:search_coarse(search_params)
					if assault_path then
						self:_merge_coarse_path_by_area(assault_path)
						assault_area = search_area
						break
					end
				end
			else
				for other_area_id, other_area in pairs(search_area.neighbours) do
					if not found_areas[other_area] then
						table.insert(to_search_areas, other_area)
						found_areas[other_area] = search_area
					end
				end
			end
		until #to_search_areas == 0
		if not assault_area and alternate_assault_area then
			assault_area = alternate_assault_area
			found_areas[assault_area] = alternate_assault_area_from
			assault_path = alternate_assault_path
		end
		if assault_area and assault_path then
			local assault_area = push and assault_area or found_areas[assault_area] == "init" and objective_area or found_areas[assault_area]
			if #assault_path > 2 and assault_area.nav_segs[assault_path[#assault_path - 1][1]] then
				table.remove(assault_path)
			end
			local used_grenade
			if push then
				local detonate_pos
				if charge then
					for c_key, c_data in pairs(assault_area.criminal.units) do
						detonate_pos = c_data.unit:movement():m_pos()
						break
					end
				else
					detonate_pos = UnitPos()
				end
				local first_chk = self._chk_group_use_flash_grenade
				local second_chk = first_chk == self._chk_group_use_flash_grenade and self._chk_group_use_flash_grenade
				used_grenade = first_chk(self, group, self._task_data.assault, detonate_pos)
				used_grenade = used_grenade or second_chk(self, group, self._task_data.assault, detonate_pos)
				self:_voice_move_in_start(group)
			end
			if not push or used_grenade then
				local grp_objective = {
					type = "assault_area",
					area = assault_area,
					coarse_path = assault_path,
					pose = "stand",
					stance = "hos",
					attitude = push and "avoid",
					moving_in = push and true,
					open_fire = push or nil,
					pushed = push or nil,
					charge = charge,
					interrupt_dis = 0
				}
				group.is_chasing = group.is_chasing or push
				self:_set_objective_to_enemy_group(group, grp_objective)
			end
		end
	elseif pull_back then
		local retreat_area, do_not_retreat
		for u_key, u_data in pairs(group.units) do
			local nav_seg_id = u_data.tracker:nav_segment()
			if current_objective.area.nav_segs[nav_seg_id] then
				retreat_area = current_objective.area
			elseif self:is_nav_seg_safe(nav_seg_id) then
				retreat_area = self:get_area_from_nav_seg_id(nav_seg_id)
			else
			end
		end
		if not retreat_area and not do_not_retreat and current_objective.coarse_path then
			local forwardmost_i_nav_point = self:_get_group_forwardmost_coarse_path_index(group)
			if forwardmost_i_nav_point then
				local nearest_safe_nav_seg_id = current_objective.coarse_path(forwardmost_i_nav_point)
				retreat_area = self:get_area_from_nav_seg_id(nearest_safe_nav_seg_id)
			end
		end
		if retreat_area then
			local new_grp_objective = {
				type = "assault_area",
				area = retreat_area,
				coarse_path = {
					{
						retreat_area.pos_nav_seg,
						mvector3.copy(retreat_area.pos)
					}
				},
				attitude = "avoid",
				pose = "crouch",
				stance = "hos"
			}
			group.is_chasing = nil
			self:_set_objective_to_enemy_group(group, new_grp_objective)
			return
		end
	end
end

function GroupAIStateBesiege:_upd_group_spawning()
	local spawn_task = self._spawning_groups[1]
	if not spawn_task then
		return
	end
	local nr_units_spawned = 0
	local produce_data = {
		name = true,
		spawn_ai = {}
	}
	local group_ai_tweak = tweak_data.group_ai				
	local spawn_points = spawn_task.spawn_group.spawn_pts
	local function _try_spawn_unit(u_type_name, spawn_entry)
		local hopeless = true
		local current_unit_type = tweak_data.levels:get_ai_group_type()
		for _, sp_data in ipairs(spawn_points) do
			local category = group_ai_tweak.unit_categories[u_type_name]
			if (sp_data.accessibility == "any" or category.access[sp_data.accessibility]) and (not sp_data.amount or sp_data.amount > 0) and sp_data.mission_element:enabled() then
				hopeless = false
				if self._t > sp_data.delay_t then
					local units = category.unit_types[current_unit_type]
					produce_data.name = units[math.random(#units)]
					local spawned_unit = sp_data.mission_element:produce(produce_data)
					local u_key = spawned_unit:key()
					local objective
					if spawn_task.objective then
						objective = self.clone_objective(spawn_task.objective)
					else
						objective = spawn_task.group.objective.element:get_random_SO(spawned_unit)
						if not objective then
							spawned_unit:set_slot(0)
							return true
						end
						objective.grp_objective = spawn_task.group.objective
					end
					local u_data = self._police[u_key]
					self:set_enemy_assigned(objective.area, u_key)
					if spawn_entry.tactics then
						u_data.tactics = spawn_entry.tactics
						u_data.tactics_map = {}
						for _, tactic_name in ipairs(u_data.tactics) do
							u_data.tactics_map[tactic_name] = true
						end
					end
					spawned_unit:brain():set_spawn_entry(spawn_entry, u_data.tactics_map)
					u_data.rank = spawn_entry.rank
					self:_add_group_member(spawn_task.group, u_key)
					if spawned_unit:brain():is_available_for_assignment(objective) then
						if objective.element then
							objective.element:clbk_objective_administered(spawned_unit)
						end
						spawned_unit:brain():set_objective(objective)
					else
						spawned_unit:brain():set_followup_objective(objective)
					end
					nr_units_spawned = nr_units_spawned + 1
					if spawn_task.ai_task then
						spawn_task.ai_task.force_spawned = spawn_task.ai_task.force_spawned + 1
					end
					sp_data.delay_t = self._t + 0.01
					if sp_data.amount then
						sp_data.amount = sp_data.amount - 1
					end
					return true
				end
			end
		end
		if hopeless then
			debug_pause("[GroupAIStateBesiege:_upd_group_spawning] spawn group", spawn_task.spawn_group.id, "failed to spawn unit", u_type_name)
			return true
		end
	end
	for u_type_name, spawn_info in pairs(spawn_task.units_remaining) do
		if not group_ai_tweak.unit_categories[u_type_name].access.acrobatic then
			for i = spawn_info.amount, 1, -1 do
				local success = _try_spawn_unit(u_type_name, spawn_info.spawn_entry)
				if success then
					spawn_info.amount = spawn_info.amount - 1
				end
			end
		end
	end
	for u_type_name, spawn_info in pairs(spawn_task.units_remaining) do
		for i = spawn_info.amount, 1, -1 do
			local success = _try_spawn_unit(u_type_name, spawn_info.spawn_entry)
			if success then
				spawn_info.amount = spawn_info.amount - 1
			end
		end
	end
	local complete = true
	for u_type_name, spawn_info in pairs(spawn_task.units_remaining) do
		if 0 < spawn_info.amount then
			complete = false
		else
		end
	end
	if complete then
		spawn_task.group.has_spawned = true
		table.remove(self._spawning_groups, 1)
		if 0 >= spawn_task.group.size then
			self._groups[spawn_task.group.id] = nil
		end
	end
end

function GroupAIStateBesiege:_chk_group_use_smoke_grenade(group, task_data, detonate_pos)
	if task_data.use_smoke and not self:is_smoke_grenade_active() then
		local shooter_pos, shooter_u_data
		local duration = 30
		for u_key, u_data in pairs(group.units) do
			if u_data.tactics_map and u_data.tactics_map.smoke_grenade then
				if not detonate_pos then
					local nav_seg_id = u_data.tracker:nav_segment()
					local nav_seg = managers.navigation._nav_segments[nav_seg_id]
					for neighbour_nav_seg_id, door_list in pairs(nav_seg.neighbours) do
						local area = self:get_area_from_nav_seg_id(neighbour_nav_seg_id)
						if task_data.target_areas[1].nav_segs[neighbour_nav_seg_id] or next(area.criminal.units) then
							local random_door_id = door_list[math.random(#door_list)]
							if type(random_door_id) == "number" then
								detonate_pos = managers.navigation._room_doors[random_door_id].center
							else
								detonate_pos = random_door_id:script_data().element:nav_link_end_pos()
							end
							shooter_pos = mvector3.copy(u_data.m_pos)
							shooter_u_data = u_data
						else
						end
					end
				end
				if detonate_pos and shooter_u_data then
					self:detonate_smoke_grenade(detonate_pos, shooter_pos, duration, false)
					task_data.use_smoke_timer = self._t + math.lerp(tweak_data.group_ai.smoke_and_flash_grenade_timeout[1], tweak_data.group_ai.smoke_and_flash_grenade_timeout[2], math.rand(0, 1) ^ 0.5)
					task_data.use_smoke = false
					if shooter_u_data.char_tweak.chatter.smoke and not shooter_u_data.unit:sound():speaking(self._t) then
						self:chk_say_enemy_chatter(shooter_u_data.unit, shooter_u_data.m_pos, "smoke")
					end
					return true
				end
			end
		end
	end
end

function GroupAIStateBesiege:_chk_group_use_flash_grenade(group, task_data, detonate_pos)
	if task_data.use_smoke and not self:is_flash_grenade_active() then
		local shooter_pos, shooter_u_data
		local duration = 3.5
		for u_key, u_data in pairs(group.units) do
			if u_data.tactics_map and u_data.tactics_map.flash_grenade then
				if not detonate_pos then
					local nav_seg_id = u_data.tracker:nav_segment()
					local nav_seg = managers.navigation._nav_segments[nav_seg_id]
					for neighbour_nav_seg_id, door_list in pairs(nav_seg.neighbours) do
						if task_data.target_areas[1].nav_segs[neighbour_nav_seg_id] then
							local random_door_id = door_list[math.random(#door_list)]
							if type(random_door_id) == "number" then
								detonate_pos = managers.navigation._room_doors[random_door_id].center
							else
								detonate_pos = random_door_id:script_data().element:nav_link_end_pos()
							end
							shooter_pos = mvector3.copy(u_data.m_pos)
							shooter_u_data = u_data
						else
						end
					end
				end
				if detonate_pos and shooter_u_data then
					self:detonate_smoke_grenade(detonate_pos, shooter_pos, duration, true)
					task_data.use_smoke_timer = self._t + tweak_data.group_ai.flash_grenade_lifetime
					task_data.use_smoke = false
					if shooter_u_data.char_tweak.chatter.flash_grenade and not shooter_u_data.unit:sound():speaking(self._t) then
						self:chk_say_enemy_chatter(shooter_u_data.unit, shooter_u_data.m_pos, "flash_grenade")
					end
					return true
				end
			end
		end
	end
end

function GroupAIStateBesiege:is_flash_grenade_active()
	return self._flash_end_t and Application:time() < self._flash_end_t
end

function GroupAIStateBesiege:_chk_group_use_cs_grenade(group, task_data, detonate_pos)
	if task_data.use_smoke and not self:is_cs_grenade_active() then
		local shooter_pos, shooter_u_data
		local duration = math.lerp((10 * 1.5), (10 * 3), math.random())
		for u_key, u_data in pairs(group.units) do
			if not detonate_pos then
				local nav_seg_id = u_data.tracker:nav_segment()
				local nav_seg = managers.navigation._nav_segments[nav_seg_id]
				for neighbour_nav_seg_id, door_list in pairs(nav_seg.neighbours) do
					if task_data.target_areas[1].nav_segs[neighbour_nav_seg_id] then
						local random_door_id = door_list[math.random(#door_list)]
						detonate_pos = UnitPos()
						shooter_pos = mvector3.copy(u_data.m_pos)
						shooter_u_data = u_data
					else
					end
				end
			end
			if detonate_pos and shooter_u_data then
				self:detonate_cs_grenade(detonate_pos, shooter_pos, duration)
				local time_window
				if self._dead_civs >= 3 then
					time_window = duration * math.lerp(tweak_data.group_ai.smoke_grenade_lifetime[1], tweak_data.group_ai.smoke_grenade_lifetime[2], math.random()) * 0.9
				elseif self._dead_civs < 3 then
					time_window = duration * math.lerp(tweak_data.group_ai.smoke_grenade_lifetime[1], tweak_data.group_ai.smoke_grenade_lifetime[2], math.random())
				end
				task_data.use_smoke_timer = self._t + 3
				task_data.use_smoke = false
				if shooter_u_data.char_tweak.chatter.smoke and not shooter_u_data.unit:sound():speaking(self._t) then
					self:chk_say_enemy_chatter(shooter_u_data.unit, shooter_u_data.m_pos, "smoke")
				end
				return true
			end
		end
	end
end

function GroupAIStateBesiege:_queue_police_upd_task()
	self._police_upd_task_queued = true
	managers.enemy:queue_task("GroupAIStateBesiege._upd_police_activity", self._upd_police_activity, self, self._t + (next(self._spawning_groups) and 0.8 or 0.5))
end

function GroupAIStateBesiege:_change_spawn_cap_pressure(amount)
	self._spawn_cap_pressure = self._spawn_cap_pressure + amount
end

function GroupAIStateBesiege:_increase_by_civilian_damage(amount)
	self._dead_civs = self._dead_civs + 1
	self._dead_civs_unlimited = self._dead_civs_unlimited + amount
	if self._dead_civs > 4 then
		self._dead_civs = 4
	end
	if self._hostage_headcount > 0 and self._dead_civs > 1 then
		self._dead_civs = self._dead_civs * 0.5
	end
end

function GroupAIStateBesiege:_civ_dead_thingy()
	return self._dead_civs > 3 and true or false 
end


function GroupAIStateBesiege:_increase_force_pool_by_camera(increase_amount)
	self._dead_cameras = self._dead_cameras + increase_amount
end

function GroupAIStateBesiege:_upd_assault_task()
	local task_data = self._task_data.assault
	if not task_data.active then
		return
	end
	local t = self._t
	local time_since_assault_started = 0
	local force_pool = (self:_get_difficulty_dependent_value(tweak_data.group_ai.besiege.assault.force_pool) * self:_get_balancing_multiplier(tweak_data.group_ai.besiege.assault.force_pool_balance_mul)) + (self._dead_civs_unlimited * (math.lerp(1, 1.5, math.random()))) - self._dead_cameras or 150 + (self._dead_civs_unlimited * (math.lerp(1, 1.5, math.random()))) - self._dead_cameras
	local task_spawn_allowance = force_pool - (self._hunt_mode and 0 or task_data.force_spawned)
	if time_since_assault_started > 0 and not self._hunt_mode then
		if task_spawn_allowance < (force_pool * 0.75) and task_spawn_allowance > (force_pool * 0.5) then
			if time_since_assault_started <= 30 then	
				self._needing_for_special_units = 4
			elseif time_since_assault_started > 30 then
				self._needing_for_special_units = 3
			elseif time_since_assault_started > 60 then
				self._needing_for_special_units = 2
			elseif time_since_assault_started > 90 then
				self._needing_for_special_units = 1
			else
				self._needing_for_special_units = 0
			end
		elseif task_spawn_allowance <= (force_pool * 0.5) and task_spawn_allowance > (force_pool * 0.25) then
			if time_since_assault_started <= 60 then
				self._needing_for_special_units = 4
			elseif time_since_assault_started > 90 then
				self._needing_for_special_units = 3
			elseif time_since_assault_started > 120 then
				self._needing_for_special_units = 2
			elseif time_since_assault_started > 180 then
				self._needing_for_special_units = 1
			else
				self._needing_for_special_units = 0
			end
		elseif task_spawn_allowance < (force_pool * 0.25) then
			if time_since_assault_started <= 150 then
				self._needing_for_special_units = 4
			elseif time_since_assault_started > 150 then
				self._needing_for_special_units = 3
			elseif time_since_assault_started > 210 then
				self._needing_for_special_units = 2
			elseif time_since_assault_started > 270 then
				self._needing_for_special_units = 1
			else
				self._needing_for_special_units = 0
			end
		end
	elseif self._hunt_mode then
		self._needing_for_special_units = 3
	end
	if task_data.phase == "anticipation" then
		if task_spawn_allowance <= 0 then
			task_data.phase = "fade"
			task_data.phase_end_t = t + tweak_data.group_ai.besiege.assault.fade_duration
		elseif t > task_data.phase_end_t then
			self._assault_number = self._assault_number + 1
			managers.mission:call_global_event("start_assault")
			managers.hud:start_assault(self._assault_number)
			self:_set_rescue_state(false)
			task_data.phase = "build"
			time_since_assault_started = self._t
			task_data.phase_end_t = self._t + 1
			task_data.is_hesitating = nil
			self:set_assault_mode(true)
			managers.trade:set_trade_countdown(false)
		else
			managers.hud:check_anticipation_voice(task_data.phase_end_t - t)
			managers.hud:check_start_anticipation_music(task_data.phase_end_t - t)
		end
	elseif task_data.phase == "build" then
		if task_spawn_allowance <= 0 then
			task_data.phase = "fade"
			task_data.phase_end_t = t + self._assault_number == 0 and 1 or tweak_data.group_ai.besiege.assault.fade_duration
		elseif t > task_data.phase_end_t or self._drama_data.zone == "high" then
			task_data.phase = "sustain"
			task_data.phase_end_t = t + math.lerp(self:_get_difficulty_dependent_value(tweak_data.group_ai.besiege.assault.sustain_duration_min), self:_get_difficulty_dependent_value(tweak_data.group_ai.besiege.assault.sustain_duration_max), math.random()) * self:_get_balancing_multiplier(tweak_data.group_ai.besiege.assault.sustain_duration_balance_mul)
		end
	elseif task_data.phase == "sustain" then
		if task_spawn_allowance <= 0 then
			task_data.phase = "fade"
			task_data.phase_end_t = t + 1
			time_since_assault_started = 0
		elseif t > task_data.phase_end_t and not self._hunt_mode then
			local hostage_delay
			if self._hostage_headcount == 1 then
				hostage_delay = 5
			elseif self._hostage_headcount == 2 then
				hostage_delay = 10
			elseif self._hostage_headcount >= 3 then
				hostage_delay = 15
			else
				hostage_delay = 0
			end
			task_data.phase = "fade"
			task_data.phase_end_t = t + tweak_data.group_ai.besiege.assault.fade_duration + hostage_delay
		end
	else
		local end_assault = false
		local enemies_left = self:_count_police_force("assault")
		if not self._hunt_mode then
			if t > task_data.phase_end_t or task_spawn_allowance <= 20 and enemies_left < 15 then
				task_data.said_retreat = true
				self:_police_announce_retreat()
				end_assault = true
			end
			if task_data.force_end or end_assault then
				print("assault task clear")
				task_data.active = nil
				task_data.phase = nil
				task_data.said_retreat = nil
				task_data.force_end = nil
				if self._draw_drama then
					self._draw_drama.assault_hist[#self._draw_drama.assault_hist][2] = t
				end
				managers.mission:call_global_event("end_assault")
				self:_begin_regroup_task()
				return
			end
		else
		end
	end
	if self._drama_data.amount <= tweak_data.drama.low then
		for criminal_key, criminal_data in pairs(self._player_criminals) do
			self:criminal_spotted(criminal_data.unit)
			for group_id, group in pairs(self._groups) do
				if group.objective.charge then
					for u_key, u_data in pairs(group.units) do
						u_data.unit:brain():clbk_group_member_attention_identified(nil, criminal_key)
					end
				end
			end
		end
	end
	local primary_target_area = task_data.target_areas[1]
	local target_pos = primary_target_area.pos
	if 1 < 2 then
		local nearest_area, nearest_dis
		for criminal_key, criminal_data in pairs(self._player_criminals) do
			if not criminal_data.status then
				local dis = mvector3.distance_sq(target_pos, criminal_data.m_pos)
				target_pos = criminal_data.m_pos
				nearest_dis = dis
				nearest_area = self:get_area_from_nav_seg_id(criminal_data.tracker:nav_segment())
			end
		end
		if nearest_area then
			primary_target_area = nearest_area
			task_data.target_areas[1] = nearest_area
		end
	end
	local nr_wanted = (task_data.force + self._dead_civs + self._spawn_cap_pressure) - self:_count_police_force("assault")
	if self._hostage_headcount > 0 then
		nr_wanted = nr_wanted * 0.85
	end
	if nr_wanted > 0 and task_data.phase == "sustain" and task_spawn_allowance > 0 then
		local used_event
		if task_data.use_spawn_event and task_data.phase ~= "anticipation" then
			task_data.use_spawn_event = false
			if self:_try_use_task_spawn_event(t, primary_target_area, "assault") then
				used_event = true
			end
		end
		if used_event then
		else
			local spawn_group, spawn_group_type = self:_find_spawn_group_near_area(primary_target_area, self._tweak_data.assault.groups, nil, nil, nil)
			if spawn_group then
				local grp_objective = {
					type = "assault_area",
					area = spawn_group.area,
					coarse_path = {
						{
							spawn_group.area.pos_nav_seg,
							spawn_group.area.pos
						}
					},
					attitude = "avoid",
					pose = "stand",
					stance = "hos"
				}
				self:_spawn_in_group(spawn_group, spawn_group_type, grp_objective, task_data)
			end
		end
	end
	if time_since_assault_started > 20 and nr_wanted > 40 then
		self._rate = 0.1
	elseif time_since_assault_started > 20 and nr_wanted > 30 then
		self._rate = 0.25
	elseif time_since_assault_started > 20 and nr_wanted > 22 then
		self._rate = 0.5
	elseif time_since_assault_started > 20 and nr_wanted > 15 then
		self._rate = 0.75
	elseif time_since_assault_started > 20 and nr_wanted > 5 then
		self._rate = 1
	elseif time_since_assault_started < 20 and nr_wanted >= 0 then
		self._rate = 2
	end
	if task_data.phase ~= "anticipation" then
		if t > task_data.use_smoke_timer then
			task_data.use_smoke = true
		end
		if self._smoke_grenade_queued and task_data.use_smoke and not self:is_smoke_grenade_active() then
			self:detonate_smoke_grenade(self._smoke_grenade_queued[1], self._smoke_grenade_queued[1], self._smoke_grenade_queued[2], self._smoke_grenade_queued[4])
			if self._smoke_grenade_queued[3] then
				self._smoke_grenade_ignore_control = true
			end
		end
	end
	self:_assign_enemy_groups_to_assault(task_data.phase)
end

--function GroupAIStateBesiege:_upd_special_assault_tasks()
--	local nr_wanted = self._needing_for_special_units - self:_count_police_force("reenforce")
--	local nearest_area, target_pos
--	for criminal_key, criminal_data in pairs(self._player_criminals) do
--		if not criminal_data.status then
--			target_pos = criminal_data.m_pos
--			nearest_area = self:get_area_from_nav_seg_id(criminal_data.tracker:nav_segment())
--		end
--	end
--	if nr_wanted > 0 and nearest_area then	
--		local used_event
--		local used_group, spawning_groups
--		if not used_event then
--			if next(self._spawning_groups) then
--				spawning_groups = true
--			else
--				local spawn_group, spawn_group_type = self:_find_spawn_group_near_area(nearest_area, tweak_data.group_ai.besiege.reenforce.groups, nil, nil, nil)
--				if spawn_group then
--					local grp_objective = {
--						type = "recon_area",
--						area = spawn_group.area,
--						target_area = nearest_area,
--						attitude = "avoid",
--						stance = "hos",
--						pose = "stand",
--						scan = true
--					}
--					self:_spawn_in_group(spawn_group, spawn_group_type, grp_objective)
--					used_group = true
--				end
--			end
--		end
--	end
--end


function GroupAIStateBesiege:_upd_recon_tasks()
	local task_data = self._task_data.recon.tasks[1]
	self:_assign_enemy_groups_to_recon()
	if not task_data then
		return
	end
	local t = self._t
	self:_assign_assault_groups_to_retire()
	local target_pos = task_data.target_area.pos
	local nr_wanted = self:_get_difficulty_dependent_value(self._tweak_data.recon.force) - self:_count_police_force("recon")
	if nr_wanted <= 0 then
		return
	end
	local used_event, used_spawn_points, reassigned
	if task_data.use_spawn_event then
		task_data.use_spawn_event = false
		if self:_try_use_task_spawn_event(t, task_data.target_area, "recon") then
			used_event = true
		end
	end
	if not used_event then
		local used_group
		if next(self._spawning_groups) then
			used_group = true
		else
			local spawn_group, spawn_group_type = self:_find_spawn_group_near_area(task_data.target_area, self._tweak_data.recon.groups, nil, nil, callback(self, self, "_verify_anticipation_spawn_point"))
			if spawn_group then
				local grp_objective = {
					type = "recon_area",
					area = spawn_group.area,
					target_area = task_data.target_area,
					attitude = "avoid",
					stance = "hos",
					scan = true
				}
				self:_spawn_in_group(spawn_group, spawn_group_type, grp_objective)
				used_group = true
			end
		end
	end
	if used_event or used_spawn_points or reassigned then
		table.remove(self._task_data.recon.tasks, 1)
		self._task_data.recon.next_dispatch_t = t + math.ceil(self:_get_difficulty_dependent_value(self._tweak_data.recon.interval)) + math.random() * self._tweak_data.recon.interval_variation
	end
end

function GroupAIStateBesiege._create_objective_from_group_objective(grp_objective, receiving_unit)
	local objective = {grp_objective = grp_objective}
	if grp_objective.element then
		objective = grp_objective.element:get_random_SO(receiving_unit)
		if not objective then
			return
		end
		objective.grp_objective = grp_objective
		return
	elseif grp_objective.type == "defend_area" or grp_objective.type == "recon_area" or grp_objective.type == "reenforce_area" then
		objective.type = "defend_area"
		objective.stance = "hos"
		objective.pose = "stand"
		objective.scan = true
		objective.interrupt_dis = 100	
		objective.interrupt_suppression = nil
		objective.path_ahead = true
	elseif grp_objective.type == "retire" then
		objective.type = "defend_area"
		objective.stance = "hos"
		objective.pose = "stand"
		objective.scan = true
		objective.interrupt_dis = 100
	elseif grp_objective.type == "assault_area" then
		objective.type = "defend_area"
		if grp_objective.follow_unit then
			objective.follow_unit = grp_objective.follow_unit
			objective.distance = grp_objective.distance
		end
		objective.stance = "hos"
		objective.pose = "stand"
		objective.scan = true
		objective.interrupt_dis = 100
		objective.interrupt_suppression = nil
		objective.path_ahead = true
	elseif grp_objective.type == "create_phalanx" then
		objective.type = "phalanx"
		objective.stance = "hos"
		objective.interrupt_dis = nil
		objective.interrupt_health = nil
		objective.interrupt_suppression = nil
		objective.attitude = "avoid"
		objective.path_ahead = true
	elseif grp_objective.type == "hunt" then
		objective.type = "hunt"
		objective.stance = "hos"
		objective.scan = true
		objective.interrupt_dis = 100
		objective.path_ahead = true
	end
	objective.stance = grp_objective.stance or objective.stance
	objective.pose = grp_objective.pose or objective.pose
	objective.area = grp_objective.area
	objective.nav_seg = grp_objective.nav_seg or objective.area.pos_nav_seg
	objective.attitude = grp_objective.attitude or objective.attitude
	objective.interrupt_dis = grp_objective.interrupt_dis or objective.interrupt_dis
	objective.interrupt_health = grp_objective.interrupt_health or objective.interrupt_health
	objective.interrupt_suppression = grp_objective.interrupt_suppression or objective.interrupt_suppression
	objective.pos = grp_objective.pos
	if grp_objective.scan ~= nil then
		objective.scan = grp_objective.scan
	end
	if grp_objective.coarse_path then
		objective.path_style = "coarse_complete"
		objective.path_data = grp_objective.coarse_path
	end
	return objective
end