local mvec3_set = mvector3.set
local mvec3_set_z = mvector3.set_z
local mvec3_sub = mvector3.subtract
local mvec3_dir = mvector3.direction
local mvec3_dot = mvector3.dot
local mvec3_dis = mvector3.distance
local mvec3_dis_sq = mvector3.distance_sq
local mvec3_lerp = mvector3.lerp
local mvec3_norm = mvector3.normalize
local temp_vec1 = Vector3()
local temp_vec2 = Vector3()
local temp_vec3 = Vector3()
CopLogicAttack.on_alert = CopLogicIdle.on_alert
CopLogicAttack.on_intimidated = CopLogicIdle.on_intimidated
function CopLogicAttack.enter(data, new_logic_name, enter_params)
	CopLogicBase.enter(data, new_logic_name, enter_params)
	data.unit:brain():cancel_all_pathing_searches()
	local old_internal_data = data.internal_data
	local my_data = {
		unit = data.unit
	}
	data.internal_data = my_data
	my_data.detection = data.char_tweak.detection.combat
	if old_internal_data then
		my_data.turning = old_internal_data.turning
		my_data.firing = old_internal_data.firing
		my_data.shooting = old_internal_data.shooting
		my_data.attention_unit = old_internal_data.attention_unit
		CopLogicAttack._set_best_cover(data, my_data, old_internal_data.best_cover)
	end
	my_data.cover_test_step = 1
	local key_str = tostring(data.key)
	my_data.detection_task_key = "CopLogicAttack._upd_enemy_detection" .. key_str
	CopLogicBase.queue_task(my_data, my_data.detection_task_key, CopLogicAttack._upd_enemy_detection, data, data.t)
	CopLogicIdle._chk_has_old_action(data, my_data)
	my_data.attitude = data.objective and data.objective.attitude or "avoid"
	my_data.weapon_range = data.char_tweak.weapon[data.unit:inventory():equipped_unit():base():weapon_tweak_data().usage].range
	data.unit:brain():set_update_enabled_state(true)
	if data.cool then
		data.unit:movement():set_cool(false)
	end
	if (not data.objective or not data.objective.stance) and data.unit:movement():stance_code() == 1 then
		data.unit:movement():set_stance("hos")
	end
	if my_data ~= data.internal_data then
		return
	end
	if data.objective and (data.objective.action_duration or data.objective.action_timeout_t and data.objective.action_timeout_t > data.t) then
		my_data.action_timeout_clbk_id = "CopLogicIdle_action_timeout" .. tostring(data.key)
		local action_timeout_t = data.objective.action_timeout_t or data.t + data.objective.action_duration
		data.objective.action_timeout_t = action_timeout_t
		CopLogicBase.add_delayed_clbk(my_data, my_data.action_timeout_clbk_id, callback(CopLogicIdle, CopLogicIdle, "clbk_action_timeout", data), action_timeout_t)
	end
	data.unit:brain():set_attention_settings({cbt = true})
end
function CopLogicAttack.exit(data, new_logic_name, enter_params)
	CopLogicBase.exit(data, new_logic_name, enter_params)
	local my_data = data.internal_data
	data.unit:brain():cancel_all_pathing_searches()
	CopLogicAttack._cancel_chase_attempt(data, my_data)
	CopLogicBase.cancel_queued_tasks(my_data)
	CopLogicBase.cancel_delayed_clbks(my_data)
	if my_data.best_cover then
		managers.navigation:release_cover(my_data.best_cover[1])
	end
	data.brain:rem_pos_rsrv("path")
	data.unit:brain():set_update_enabled_state(true)
end

function CopLogicAttack._cancel_chase_attempt(data, my_data)
	my_data.chase_path = nil
	if my_data.walking_to_chase_pos then
		local new_action = {type = "idle", body_part = 2}
		data.unit:brain():action_request(new_action)
	elseif my_data.pathing_to_chase_pos then
		data.brain:rem_pos_rsrv("path")
		if data.active_searches[my_data.chase_path_search_id] then
			managers.navigation:cancel_pathing_search(my_data.chase_path_search_id)
			data.active_searches[my_data.chase_path_search_id] = nil
		elseif data.pathing_results then
			data.pathing_results[my_data.chase_path_search_id] = nil
		end
		my_data.chase_path_search_id = nil
		my_data.pathing_to_chase_pos = nil
		data.unit:brain():cancel_all_pathing_searches()
	elseif my_data.chase_pos then
		my_data.chase_pos = nil
	end
end

function CopLogicAttack.is_advancing(data)
	if data.internal_data.moving_to_cover then
		return data.internal_data.moving_to_cover[1][1]
	end
	if data.internal_data.walking_to_cover_shoot_pos then
		return data.internal_data.walking_to_cover_shoot_pos._last_pos
	end
	if data.internal_data.walking_to_chase_pos and data.pos_rsrv.move_dest then
		return data.pos_rsrv.move_dest.position
	end	
end

function CopLogicAttack.update(data)
	local my_data = data.internal_data
	local unit = data.unit
	if my_data.has_old_action then
		CopLogicAttack._upd_stop_old_action(data, my_data)
		if not my_data.update_queue_id then
			data.unit:brain():set_update_enabled_state(false)
			my_data.update_queue_id = "CopLogicAttack.queued_update" .. tostring(data.key)
			CopLogicAttack.queue_update(data, my_data)
		end
		return
	end
	if CopLogicIdle._chk_relocate(data) then
		return
	end
	if CopLogicAttack._chk_exit_non_walkable_area(data) then
		return
	end
	if not data.attention_obj or data.attention_obj.reaction < AIAttentionObject.REACT_AIM then
		CopLogicAttack._upd_enemy_detection(data, true)
		if my_data ~= data.internal_data or not data.attention_obj then
			return
		end
	end
	if data.team.id == "criminal1" and (not data.objective or data.objective.type == "free") and (not data.path_fail_t or data.t - data.path_fail_t > 3) then
		managers.groupai:state():on_criminal_jobless(data.unit)
		if my_data ~= data.internal_data then
			return
		end
	end
	my_data.want_to_take_cover = true
	CopLogicAttack._update_cover(data)
	CopLogicAttack._process_pathing_results(data, my_data)
	CopLogicAttack._upd_combat_movement(data)
	if not my_data.update_queue_id then
		data.unit:brain():set_update_enabled_state(false)
		my_data.update_queue_id = "CopLogicAttack.queued_update" .. tostring(data.key)
		CopLogicAttack.queue_update(data, my_data)
	end
end

function CopLogicAttack._process_pathing_results(data, my_data)
	if not data.pathing_results then
		return
	end
	local pathing_results = data.pathing_results
	data.pathing_results = nil
	local path = pathing_results[my_data.cover_path_search_id]
	if path then
		if path ~= "failed" then
			my_data.cover_path = path
		else
			print(data.unit, "[CopLogicAttack._process_pathing_results] cover path failed", data.unit)
			CopLogicAttack._set_best_cover(data, my_data, nil)
			my_data.cover_path_failed_t = TimerManager:game():time()
		end
		my_data.processing_cover_path = nil
		my_data.cover_path_search_id = nil
	end
	path = pathing_results[my_data.chase_path_search_id]
	if path then
		if path ~= "failed" then
			my_data.chase_path = path
		else
			print("[CopLogicAttack._process_pathing_results] chase path failed")
		end
		my_data.pathing_to_chase_pos = nil
		my_data.chase_path_search_id = nil
	end
	path = pathing_results[my_data.charge_path_search_id]
	if path then
		if path ~= "failed" then
			my_data.charge_path = path
		else
			print("[CopLogicAttack._process_pathing_results] charge path failed", data.unit)
		end
		my_data.charge_path_search_id = nil
		my_data.charge_path_failed_t = TimerManager:game():time()
	end
	path = pathing_results[my_data.expected_pos_path_search_id]
	if path then
		if path ~= "failed" then
			my_data.expected_pos_path = path
		end
		my_data.expected_pos_path_search_id = nil
	end
end

function CopLogicAttack.action_complete_clbk(data, action)
	local my_data = data.internal_data
	local action_type = action:type()
	if action_type == "walk" then
		my_data.advancing = nil
		if my_data.walking_to_chase_pos then
			my_data.walking_to_chase_pos = nil
		end
		CopLogicAttack._cancel_cover_pathing(data, my_data)
		CopLogicAttack._cancel_charge(data, my_data)
		if my_data.surprised then
			my_data.surprised = false
		elseif my_data.moving_to_cover then
			if action:expired() then
				my_data.in_cover = my_data.moving_to_cover
				my_data.cover_enter_t = data.t
			end
			my_data.moving_to_cover = nil
		elseif my_data.walking_to_cover_shoot_pos then
			my_data.walking_to_cover_shoot_pos = nil
			my_data.at_cover_shoot_pos = true
		end
	elseif action_type == "shoot" then
		my_data.shooting = nil
	elseif action_type == "turn" then
		my_data.turning = nil
	elseif action_type == "hurt" then
		CopLogicAttack._cancel_cover_pathing(data, my_data)
		if data.important and action:expired() and not CopLogicBase.chk_start_action_dodge(data, "hit") then
			data.logic._upd_aim(data, my_data)
		end
	elseif action_type == "dodge" then
		local timeout = action:timeout()
		if timeout then
			data.dodge_timeout_t = TimerManager:game():time() + math.lerp(timeout[1], timeout[2], math.random())
		end
		CopLogicAttack._cancel_cover_pathing(data, my_data)
		if action:expired() then
			CopLogicAttack._upd_aim(data, my_data)
		end
	end
end

function CopLogicAttack.queue_update(data, my_data)
	CopLogicBase.queue_task(my_data, my_data.update_queue_id, data.logic.queued_update, data, data.t + 0.5)
end

function CopLogicAttack._upd_combat_movement(data)
	local my_data = data.internal_data
	local t = data.t
	local unit = data.unit
	local focus_enemy = data.attention_obj
	local in_cover = my_data.in_cover
	local best_cover = my_data.best_cover
	local enemy_visible = focus_enemy.verified
	local enemy_visible_soft = focus_enemy.verified_t and t - focus_enemy.verified_t < 2
	local enemy_visible_softer = focus_enemy.verified_t and t - focus_enemy.verified_t < 15
	local alert_soft = data.is_suppressed
	local action_taken = data.logic.action_taken(data, my_data)
	local attitude = my_data.attitude
	local range = 1
	local max_dist = 400
	local want_to_take_cover = true
	action_taken = action_taken or CopLogicAttack._upd_pose(data, my_data)
	local move_to_cover 
	local want_flank_cover = true
	if my_data.cover_test_step ~= 1 and not enemy_visible_softer and (action_taken or want_to_take_cover or not in_cover) then
		my_data.cover_test_step = 1
	end
	if my_data.stay_out_time and (enemy_visible_soft or not my_data.at_cover_shoot_pos or action_taken or want_to_take_cover) then
		my_data.stay_out_time = nil
	elseif my_data.attitude == "engage" and not my_data.stay_out_time and not enemy_visible_soft and my_data.at_cover_shoot_pos and not action_taken and not want_to_take_cover then
		my_data.stay_out_time = nil
	end
	if data.tactics and data.tactics.charge and data.objective and data.objective.grp_objective and data.objective.grp_objective.charge and (not my_data.charge_path_failed_t or data.t - my_data.charge_path_failed_t > 3) then
		if not my_data.charge_path_search_id and data.attention_obj.nav_tracker then
			my_data.charge_pos = CopLogicTravel._get_pos_on_wall(data.attention_obj.nav_tracker:field_position(), max_dist, 45, nil)
			if my_data.charge_pos then
				my_data.charge_path_search_id = "charge" .. tostring(data.key)
				unit:brain():search_for_path(my_data.charge_path_search_id, my_data.charge_pos, nil, nil, nil)
			else
				debug_pause_unit(data.unit, "failed to find charge_pos", data.unit)
				my_data.charge_path_failed_t = TimerManager:game():time()
			end
		end
	else
		local my_tracker = unit:movement():nav_tracker()
		move_to_cover = true
	end
	if not my_data.processing_cover_path and not my_data.cover_path and not my_data.charge_path_search_id and not action_taken and best_cover and (not in_cover or best_cover[1] ~= in_cover[1]) and (not my_data.cover_path_failed_t or data.t - my_data.cover_path_failed_t > 3) then
		CopLogicAttack._cancel_cover_pathing(data, my_data)
		local search_id = tostring(unit:key()) .. "cover"
		if data.unit:brain():search_for_path_to_cover(search_id, best_cover[1], best_cover[5]) then
			my_data.cover_path_search_id = search_id
			my_data.processing_cover_path = best_cover
		end
	end
	if my_data.walking_to_chase_pos then
	elseif my_data.pathing_to_chase_pos then
	elseif my_data.chase_pos then
		local to_pos = my_data.chase_pos
		my_data.chase_path_search_id = tostring(unit:key()) .. "chase"
		my_data.pathing_to_chase_pos = true
		my_data.chase_pos = nil
		data.brain:add_pos_rsrv("path", {
			position = mvector3.copy(to_pos),
			radius = 60
		})
		unit:brain():search_for_path(my_data.chase_path_search_id, to_pos)
	elseif focus_enemy.nav_tracker then 
		my_data.chase_pos = CopLogicAttack._find_flank_pos(data, my_data, focus_enemy.nav_tracker)
	end
	if not action_taken then
		if not my_data.chase_path and my_data.cover_path then
			action_taken = CopLogicAttack._chk_request_action_walk_to_cover(data, my_data)
		elseif not my_data.chase_path and my_data.charge_path then
			local path = my_data.charge_path
			my_data.charge_path = nil
			action_taken = CopLogicAttack._chk_request_action_walk_to_cover_shoot_pos(data, my_data, path)
		elseif my_data.chase_path then
			action_taken = CopLogicAttack._chk_request_action_walk_to_chase_pos(data, my_data)
		elseif not my_data.chase_path and not my_data.cover_path and not my_data.charge_path then
			action_taken = CopLogicAttack._chk_start_action_move_out_of_the_way(data, my_data)
		end
	end
	if focus_enemy and focus_enemy.verified and focus_enemy.dis < 650 then
		if not my_data.flank_cover then
			local sign = math.random() < 0.5 and -1 or 1
			local step = 30
			my_data.flank_cover = {
				step = step,
				angle = step * sign,
				sign = sign
			}
		end
	end
	if data.attention_obj.verified then
		if data.is_suppressed and data.t - data.unit:character_damage():last_suppression_t() < 0.7 and (not data.unit:movement():chk_action_forbidden("walk") or focus_enemy.dis <= 400) then
			action_taken = CopLogicBase.chk_start_action_dodge(data, "scared")
		end
		if not action_taken and focus_enemy.is_person and focus_enemy.dis <= 850 or focus_enemy.dis <= 400 and focus_enemy.is_person then
			if focus_enemy.aimed_at or not data.unit:movement():chk_action_forbidden("walk") or math.random() <= 0.55 then
				action_taken = math.random() < 0.75 and CopLogicBase.chk_start_action_dodge(data, "hit") or CopLogicBase.chk_start_action_dodge(data, "scared")
			end
		end
	end
	local grenade = true
	local in_idle_t = 0
	local overloaded = data.t
	if focus_enemy.is_human_player then
		if focus_enemy.is_local_player then
			local e_movement_state = focus_enemy.unit:movement():current_state()
			if e_movement_state:_interacting() then
				grenade = true
			end
		elseif focus_enemy.is_husk_player then
			local e_anim_data = focus_enemy.unit:anim_data()
			if e_anim_data.idle then
				in_idle_t = data.t
			elseif e_anim_data.move then
				in_idle_t = 0
			end
		end
	end
	if overloaded > 20 then
		grenade = true
		overloaded = data.t
	end
	if in_idle_t > 3 then
		grenade = true
	end
	if grenade and not managers.groupai:state():is_flash_grenade_active() and managers.groupai:state():_is_assault_active() and data.attention_obj.verified and focus_enemy.dis < 850 then
		managers.groupai:state():detonate_smoke_grenade(focus_enemy.m_pos, data.unit:movement():m_head_pos(), math.lerp(3.5, 5, math.random()), true)
	elseif grenade and not managers.groupai:state():is_smoke_grenade_active() and managers.groupai:state():_is_assault_active() and data.attention_obj.verified and focus_enemy.dis < 1400 then
		managers.groupai:state():detonate_smoke_grenade(focus_enemy.m_pos, data.unit:movement():m_head_pos(), math.lerp(8, 15, math.random()))
	end
end

function CopLogicAttack._chk_request_action_walk_to_chase_pos(data, my_data, end_rot)
	CopLogicAttack._correct_path_start_pos(data, my_data.chase_path)
	if not data.unit:movement():chk_action_forbidden("walk") then
		local new_action_data = {
			type = "walk",
			nav_path = my_data.chase_path,
			variant = "run",
			body_part = 2,
			end_rot = end_rot
		}
		my_data.chase_path = nil
		my_data.walking_to_chase_pos = data.unit:brain():action_request(new_action_data)
		if my_data.walking_to_chase_pos then
			data.brain:rem_pos_rsrv("path")
		end
	end
end

function CopLogicAttack._chk_start_action_move_out_of_the_way(data, my_data, focus_enemy)
	local my_tracker = data.unit:movement():nav_tracker()
	local reservation = {
		position = data.m_pos,
		radius = 20,
		filter = data.pos_rsrv_id
	}
	if not managers.navigation:is_pos_free(reservation) then
		local to_pos = CopLogicTravel._get_pos_on_wall(data.m_pos, 200)
		if to_pos then
			local path = {
				my_tracker:position(),
				to_pos
			}
			CopLogicAttack._chk_request_action_walk_to_cover_shoot_pos(data, my_data, path, "run")
		end
	end
end


function CopLogicAttack._chk_request_action_walk_to_cover(data, my_data)
	CopLogicAttack._correct_path_start_pos(data, my_data.cover_path)
	local new_action_data = {
		type = "walk",
		nav_path = my_data.cover_path,
		variant = "run",
		body_part = 2,
		end_pose = "stand"
	}
	my_data.cover_path = nil
	my_data.advancing = data.unit:brain():action_request(new_action_data)
	if my_data.advancing then	
		my_data.walking_to_cover_shoot_pos = my_data.advancing
		my_data.at_cover_shoot_pos = nil
		my_data.in_cover = nil
		data.brain:rem_pos_rsrv("path")
	end
end

function CopLogicAttack._chk_request_action_walk_to_cover_shoot_pos(data, my_data, path, speed)
	CopLogicAttack._correct_path_start_pos(data, path)
	CopLogicAttack._cancel_charge(data, my_data)
	local new_action_data = {
		type = "walk",
		nav_path = path,
		variant = "run",
		body_part = 2,
		end_pose = "stand"
	}
	my_data.cover_path = nil
	my_data.advancing = data.unit:brain():action_request(new_action_data)
	if my_data.advancing then
		my_data.walking_to_cover_shoot_pos = my_data.advancing
		my_data.at_cover_shoot_pos = nil
		my_data.in_cover = nil
		data.brain:rem_pos_rsrv("path")
	end
end

function CopLogicAttack._update_cover(data)
	local my_data = data.internal_data
	local cover_release_dis_sq = nil
	local best_cover = my_data.best_cover
	local attitude = my_data.attitude
	local satisfied = true
	local my_pos = data.m_pos
	if data.attention_obj and data.attention_obj.nav_tracker and data.attention_obj.reaction >= AIAttentionObject.REACT_COMBAT then
		local find_new = not my_data.moving_to_cover and not my_data.walking_to_cover_shoot_pos
		if find_new then
			local enemy_tracker = data.attention_obj.nav_tracker
			local threat_pos = enemy_tracker:field_position()
			if data.objective and data.objective.type == "follow" then
				local near_pos = data.objective.follow_unit:movement():m_pos()
				if (not best_cover or not CopLogicAttack._verify_follow_cover(best_cover[1], near_pos, threat_pos, nil, 400)) and not my_data.processing_cover_path and not my_data.charge_path_search_id then
					local follow_unit_area = managers.groupai:state():get_area_from_nav_seg_id(data.objective.follow_unit:movement():nav_tracker():nav_segment())
					local found_cover = managers.navigation:find_cover_in_nav_seg_3(follow_unit_area.nav_segs, data.objective.distance and data.objective.distance * (attitude == "engage" and 0.9 or 0.3) or nil, near_pos, threat_pos)
					if found_cover then
						if not follow_unit_area.nav_segs[found_cover[3]:nav_segment()] then
							debug_pause_unit(data.unit, "cover in wrong area")
						end
						satisfied = true
						local better_cover = {found_cover}
						CopLogicAttack._set_best_cover(data, my_data, better_cover)
						local offset_pos, yaw = CopLogicAttack._get_cover_offset_pos(data, better_cover, threat_pos)
						if offset_pos then
							better_cover[5] = offset_pos
							better_cover[6] = yaw
						end
					end
				end
			else
				local want_to_take_cover = my_data.want_to_take_cover
				local flank_cover = my_data.flank_cover
				local min_dis, max_dis
				if want_to_take_cover then
					min_dis = 400
				end
				if not my_data.processing_cover_path and not my_data.charge_path_search_id and (not best_cover or flank_cover or not CopLogicAttack._verify_cover(best_cover[1], threat_pos, 0, max_dis)) then
					satisfied = false
					local my_vec = my_pos - threat_pos
					if flank_cover then
						mvector3.rotate_with(my_vec, Rotation(flank_cover.angle))
					end
					local min_dis = 0
					local my_side_pos = threat_pos + my_vec
					mvector3.set_length(my_vec, 400)
					local furthest_side_pos = threat_pos + my_vec
					if flank_cover then
						local angle = flank_cover.angle
						local sign = flank_cover.sign
						if math.sign(angle) ~= sign then
							angle = -angle + flank_cover.step * sign
							if math.abs(angle) > 90 then
								flank_cover.failed = true
							else
								flank_cover.angle = angle
							end
						else
							flank_cover.angle = -angle
						end
					end
					local min_threat_dis, cone_angle
					if flank_cover then
						cone_angle = flank_cover.step
					else
						cone_angle = math.lerp(90, 60, math.min(1, min_dis / 3000))
					end
					local found_cover = managers.navigation:find_cover_near_pos_1(threat_pos, threat_pos, 400, nil, true)
					if found_cover and (not best_cover or CopLogicAttack._verify_cover(found_cover, threat_pos, 0, max_dis)) then
						satisfied = true
						local better_cover = {found_cover}
						CopLogicAttack._set_best_cover(data, my_data, better_cover)
						local offset_pos, yaw = CopLogicAttack._get_cover_offset_pos(data, better_cover, threat_pos)
						if offset_pos then
							better_cover[5] = offset_pos
							better_cover[6] = yaw
						end
					end
				end
			end
		end
	elseif best_cover then
		CopLogicAttack._set_best_cover(data, my_data, nil)
	end
end
function CopLogicAttack._verify_cover(cover, threat_pos, min_dis, max_dis)
	local threat_dis = mvector3.direction(temp_vec1, cover[1], threat_pos)
	if 400 > threat_dis then
		return
	end
	return true
end

function CopLogicAttack._verify_new_to_pos(to_pos, threat_pos, my_pos)
	local to_pos_threat_pos = to_pos - threat_pos
	local m_pos_threat_pos = my_pos - threat_pos
	if to_pos_threat_pos > m_pos_threat_pos then
		return
	end
	return true
end

function CopLogicAttack._upd_enemy_detection(data, is_synchronous)
	managers.groupai:state():on_unit_detection_updated(data.unit)
	data.t = TimerManager:game():time()
	local my_data = data.internal_data
	local min_reaction = AIAttentionObject.REACT_COMBAT
	local new_attention, new_prio_slot, new_reaction = CopLogicIdle._get_priority_attention(data, data.detected_attention_objects, nil)
	local old_att_obj = data.attention_obj
	data.logic._set_attention_obj(data, new_attention, new_reaction)
	data.logic._chk_exit_attack_logic(data, new_reaction)
	if my_data ~= data.internal_data then
		return
	end
	CopLogicBase._chk_call_the_police(data)
	if my_data ~= data.internal_data then
		return
	end
	if my_data == data.internal_data then
		data.logic._upd_aim(data, my_data)
		if not is_synchronous then
			CopLogicBase.queue_task(my_data, my_data.detection_task_key, CopLogicAttack._upd_enemy_detection, data, data.t + 0.5)
		end
	end
	CopLogicBase._report_detections(data.detected_attention_objects)
end

function CopLogicAttack._set_attention_obj(data, new_att_obj, new_reaction)
	local old_att_obj = data.attention_obj
	data.attention_obj = new_att_obj
	if new_att_obj then
		new_reaction = new_reaction or new_att_obj.settings.reaction
		new_att_obj.stare_expire_t = nil
		new_att_obj.pause_expire_t = nil
		new_att_obj.reaction = new_reaction
		local new_crim_rec = new_att_obj.criminal_record
		local is_same_obj, contact_chatter_time_ok
		if old_att_obj then
			if old_att_obj.u_key == new_att_obj.u_key then
				is_same_obj = true
				contact_chatter_time_ok = new_crim_rec and data.t - new_crim_rec.det_t > 0.1
				new_att_obj.stare_expire_t = nil
				new_att_obj.pause_expire_t = nil
			else
				old_att_obj.stare_expire_t = nil
				old_att_obj.pause_expire_t = nil
				if old_att_obj.criminal_record then
					managers.groupai:state():on_enemy_disengaging(data.unit, old_att_obj.u_key)
				end
				if new_crim_rec then
					managers.groupai:state():on_enemy_engaging(data.unit, new_att_obj.u_key)
				end
				contact_chatter_time_ok = new_crim_rec and data.t - new_crim_rec.det_t > 0.1
			end
		else
			if new_crim_rec then
				managers.groupai:state():on_enemy_engaging(data.unit, new_att_obj.u_key)
			end
			contact_chatter_time_ok = new_crim_rec and data.t - new_crim_rec.det_t > 0.1
		end
		if not is_same_obj then
			if new_att_obj.settings.duration then
				new_att_obj.stare_expire_t = nil
				new_att_obj.pause_expire_t = nil
			end
			new_att_obj.acquire_t = data.t
		end
		if new_reaction >= AIAttentionObject.REACT_SHOOT and new_att_obj.verified and new_att_obj.is_person and data.char_tweak.chatter.contact then
			data.unit:sound():say("c01", true)
		end
	elseif old_att_obj and old_att_obj.criminal_record then
		managers.groupai:state():on_enemy_disengaging(data.unit, old_att_obj.u_key)
	end
end

function CopLogicAttack._upd_aim(data, my_data)
	local shoot, aim, expected_pos
	local focus_enemy = data.attention_obj
	if focus_enemy and focus_enemy.reaction >= AIAttentionObject.REACT_AIM then
		local last_sup_t = data.unit:character_damage():last_suppression_t()
		if focus_enemy.verified or focus_enemy.nearly_visible then
			if aim == nil and focus_enemy.reaction >= AIAttentionObject.REACT_AIM then
				if focus_enemy.reaction >= AIAttentionObject.REACT_SHOOT then
					local running = my_data.advancing and not my_data.advancing:stopping() and my_data.advancing:haste() == "run"
					if focus_enemy.verified then
						shoot = true
					end
					if not shoot or not aim or not shoot and not aim then
						if focus_enemy.dis <= 150 then
							focus_enemy.reaction = AIAttentionObject.REACT_ARREST
						end
						if focus_enemy.verified then
							shoot = true
						else
							local time_since_verification = focus_enemy.verified_t and data.t - focus_enemy.verified_t
							if my_data.firing and time_since_verification and time_since_verification < (my_data.attitude == engage and not data.team.id == "criminal1" and 7.5 or 3.5) then
								shoot = true
							else
								data.brain:search_for_path_to_unit("hunt" .. tostring(my_data.key), focus_enemy.unit)
							end
						end
					end
				else
					aim = true
				end
			end
		elseif focus_enemy.reaction >= AIAttentionObject.REACT_AIM then
			local time_since_verification = focus_enemy.verified_t and data.t - focus_enemy.verified_t
			local running = my_data.advancing and not my_data.advancing:stopping() and my_data.advancing:haste() == "run"
			local same_z = math.abs(focus_enemy.verified_pos.z - data.m_pos.z) < 250
			if running then
				if time_since_verification and time_since_verification < math.lerp(5, 1, math.max(0, focus_enemy.verified_dis - 500) / 600) then
					aim = true
				end
			else
				aim = true
			end
			if aim and my_data.shooting and focus_enemy.reaction >= AIAttentionObject.REACT_SHOOT and time_since_verification then
				if time_since_verification < (running and 2 or 3) then
					shoot = true
				end
			end
			if not aim then
				expected_pos = CopLogicAttack._get_expected_attention_position(data, my_data)
				if expected_pos then
					if running then
						local watch_dir = temp_vec1
						mvec3_set(watch_dir, expected_pos)
						mvec3_sub(watch_dir, data.m_pos)
						mvec3_set_z(watch_dir, 0)
						local watch_pos_dis = mvec3_norm(watch_dir)
						local walk_to_pos = data.unit:movement():get_walk_to_pos()
						local walk_vec = temp_vec2
						mvec3_set(walk_vec, walk_to_pos)
						mvec3_sub(walk_vec, data.m_pos)
						mvec3_set_z(walk_vec, 0)
						mvec3_norm(walk_vec)
						local watch_walk_dot = mvec3_dot(watch_dir, walk_vec)
						if watch_pos_dis < 500 or watch_pos_dis < 1000 and watch_walk_dot > 0.85 then
							aim = true
						end
					else
						aim = true
					end
				end
			end
		else
			expected_pos = CopLogicAttack._get_expected_attention_position(data, my_data)
			if expected_pos then
				aim = true
			end
		end
	end
	if not aim and data.char_tweak.always_face_enemy and focus_enemy and focus_enemy.reaction >= AIAttentionObject.REACT_COMBAT then
		if focus_enemy.verified then
			shoot = true
		else
			aim = true
		end
	end
	if data.logic.chk_should_turn(data, my_data) and (focus_enemy or expected_pos) then
		local enemy_pos = expected_pos or focus_enemy.verified and focus_enemy.m_head_pos or focus_enemy.verified_pos
		CopLogicAttack._chk_request_action_turn_to_enemy(data, my_data, data.m_pos, enemy_pos)
	end
	if aim or shoot then
		if expected_pos then
			if my_data.attention_unit ~= expected_pos then
				CopLogicBase._set_attention_on_pos(data, mvector3.copy(expected_pos))
				my_data.attention_unit = mvector3.copy(expected_pos)
			end
		elseif focus_enemy.verified or focus_enemy.nearly_visible then
			if my_data.attention_unit ~= focus_enemy.u_key then
				CopLogicBase._set_attention(data, focus_enemy)
				my_data.attention_unit = focus_enemy.u_key
			end
		else
			local look_pos = focus_enemy.last_verified_pos or focus_enemy.verified_pos
			if my_data.attention_unit ~= look_pos then
				CopLogicBase._set_attention_on_pos(data, mvector3.copy(look_pos))
				my_data.attention_unit = mvector3.copy(look_pos)
			end
		end
		if not my_data.shooting and not my_data.spooc_attack and not data.unit:anim_data().reload and not data.unit:movement():chk_action_forbidden("action") then
			local shoot_action = {type = "shoot", body_part = 3}
			if data.unit:brain():action_request(shoot_action) then
				my_data.shooting = true
			end
		end
	else
		if my_data.shooting then
			local new_action
			if data.unit:anim_data().reload then
				new_action = {type = "reload", body_part = 3}
			else
				new_action = {type = "idle", body_part = 3}
			end
			data.unit:brain():action_request(new_action)
		end
		if my_data.attention_unit then
			CopLogicBase._reset_attention(data)
			my_data.attention_unit = nil
		end
	end
	CopLogicAttack.aim_allow_fire(shoot, aim, data, my_data)
end

function CopLogicAttack.chk_should_turn(data, my_data)
	return not my_data.turning and not data.unit:movement():chk_action_forbidden("walk") and not my_data.walking_to_chase_pos
end
function CopLogicAttack._get_cover_offset_pos(data, cover_data, threat_pos)
	local threat_vec = threat_pos - cover_data[1][1]
	mvector3.set_z(threat_vec, 0)
	local threat_polar = threat_vec:to_polar_with_reference(cover_data[1][2], math.UP)
	local threat_spin = threat_polar.spin
	local rot
	if threat_spin < -20 then
		rot = Rotation(90)
	elseif threat_spin > 20 then
		rot = Rotation(-90)
	else
		rot = Rotation(180)
	end
	local offset_pos = mvector3.copy(cover_data[1][2])
	mvector3.rotate_with(offset_pos, rot)
	mvector3.set_length(offset_pos, 25)
	mvector3.add(offset_pos, cover_data[1][1])
	local ray_params = {
		tracker_from = cover_data[1][3],
		pos_to = offset_pos,
		trace = true
	}
	managers.navigation:raycast(ray_params)
	return ray_params.trace[1]
end
function CopLogicAttack._find_flank_pos(data, my_data, flank_tracker, max_dist)
	local pos = flank_tracker:position()
	local vec_to_pos = pos - data.m_pos
	mvector3.set_z(vec_to_pos, 0)
	local max_dis = 1200
	mvector3.set_length(vec_to_pos, max_dis)
	local accross_positions = managers.navigation:find_walls_accross_tracker(flank_tracker, vec_to_pos, 160, 5)
	if accross_positions then
		local optimal_dis = 0.7
		local best_error_dis, best_pos, best_is_hit, best_is_miss, best_has_too_much_error
		for _, accross_pos in ipairs(accross_positions) do
			local error_dis = math.abs(mvector3.distance(accross_pos[1], pos) - optimal_dis)
			local too_much_error = error_dis / optimal_dis > 0.2
			local is_hit = accross_pos[2]
			if best_is_hit then
				if is_hit then
					if best_error_dis > error_dis then
						local reservation = {
							position = accross_pos[1],
							radius = 30,
							filter = data.pos_rsrv_id
						}
						if managers.navigation:is_pos_free(reservation) then
							best_pos = accross_pos[1]
							best_error_dis = error_dis
							best_has_too_much_error = too_much_error
						end
					end
				elseif best_has_too_much_error then
					local reservation = {
						position = accross_pos[1],
						radius = 30,
						filter = data.pos_rsrv_id
					}
					if managers.navigation:is_pos_free(reservation) then
						best_pos = accross_pos[1]
						best_error_dis = error_dis
						best_is_miss = true
						best_is_hit = nil
					end
				end
			elseif best_is_miss then
				if not too_much_error then
					local reservation = {
						position = accross_pos[1],
						radius = 30,
						filter = data.pos_rsrv_id
					}
					if managers.navigation:is_pos_free(reservation) then
						best_pos = accross_pos[1]
						best_error_dis = error_dis
						best_has_too_much_error = nil
						best_is_miss = nil
						best_is_hit = true
					end
				end
			else
				local reservation = {
					position = accross_pos[1],
					radius = 30,
					filter = data.pos_rsrv_id
				}
				if managers.navigation:is_pos_free(reservation) then
					best_pos = accross_pos[1]
					best_is_hit = is_hit
					best_is_miss = not is_hit
					best_has_too_much_error = too_much_error
					best_error_dis = error_dis
				end
			end
		end
		return best_pos
	end
end

function CopLogicAttack.is_available_for_assignment(data, new_objective)
	local my_data = data.internal_data
	if my_data.exiting then
		return
	end
	if new_objective and new_objective.forced then
		return true
	end
	if data.unit:movement():chk_action_forbidden("walk") then
		return
	end
	if data.path_fail_t and data.t < data.path_fail_t + 3 then
		return
	end
	local att_obj = data.attention_obj
	if not att_obj or att_obj.reaction < AIAttentionObject.REACT_AIM then
		return true
	end
	if not new_objective or new_objective.type == "free" then
		return true
	end
	return true
end
function CopLogicAttack._chk_wants_to_take_cover(data, my_data)
	if not data.attention_obj or data.attention_obj.reaction < AIAttentionObject.REACT_COMBAT then
		return
	end
	return true
end
function CopLogicAttack._set_best_cover(data, my_data, cover_data)
	local best_cover = my_data.best_cover
	if best_cover then
		managers.navigation:release_cover(best_cover[1])
		CopLogicAttack._cancel_cover_pathing(data, my_data)
	end
	if cover_data then
		managers.navigation:reserve_cover(cover_data[1], data.pos_rsrv_id)
		my_data.best_cover = cover_data
		if not my_data.in_cover and not my_data.walking_to_cover_shoot_pos and not my_data.moving_to_cover and mvec3_dis_sq(cover_data[1][1], data.m_pos) < 100 then
			my_data.in_cover = my_data.best_cover
			my_data.cover_enter_t = data.t
		end
	else
		my_data.best_cover = nil
		my_data.flank_cover = nil
	end
end
function CopLogicAttack._set_nearest_cover(my_data, cover_data)
	local nearest_cover = my_data.nearest_cover
	if nearest_cover then
		managers.navigation:release_cover(nearest_cover[1])
	end
	if cover_data then
		local pos_rsrv_id = my_data.unit:movement():pos_rsrv_id()
		managers.navigation:reserve_cover(cover_data[1], pos_rsrv_id)
		my_data.nearest_cover = cover_data
	else
		my_data.nearest_cover = nil
	end
end

function CopLogicAttack._get_all_paths(data)
	return {
		cover_path = data.internal_data.cover_path,
		flank_path = data.internal_data.flank_path,
		chase_path = data.internal_data.chase_path
	}
end
function CopLogicAttack._set_verified_paths(data, verified_paths)
	data.internal_data.cover_path = verified_paths.cover_path
	data.internal_data.flank_path = verified_paths.flank_path
	data.internal_data.chase_path = verified_paths.chase_path
end
function CopLogicAttack.action_taken(data, my_data)
	return my_data.turning or my_data.has_old_action or data.unit:movement():chk_action_forbidden("walk")
end

function CopLogicAttack._find_retreat_position(from_pos, threat_pos, threat_head_pos, threat_tracker, max_dist, vis_required)
	return threat_pos
end
