function GroupAIStateBase:a_true_callback()
	return not self:is_area_safe(area)
end
function GroupAIStateBase:is_area_safe(area)
	for u_key, u_data in pairs(self._criminals) do
		if area.nav_segs[u_data.tracker:nav_segment()] then
			return
		end
	end
	return true
end
function GroupAIStateBase:is_area_safe_assault(area)
	for u_key, u_data in pairs(self._criminals) do
		if not u_data.is_deployable and area.nav_segs[u_data.tracker:nav_segment()] then
			return
		end
	end
	return true
end

function GroupAIStateBase:_clbk_switch_enemies_to_not_cool()
	for u_key, unit_data in pairs(self._police) do
		if unit_data.unit:movement():cool() and unit_data.assigned_area then
			unit_data.unit:movement():set_cool(false)
			if unit_data.unit:brain():is_available_for_assignment() then
				local new_objective
				local valid_criminals = {}
				for pl_key, pl_record in pairs(self._player_criminals) do
					if pl_record.status ~= "dead" then
						table.insert(valid_criminals, pl_key)
					end
				end
				if #valid_criminals > 0 then
					local follow_unit = self._player_criminals[valid_criminals[math.random(#valid_criminals)]].unit
					new_objective = {
						type = "follow",
						follow_unit = follow_unit,
						scan = true,
						is_default = true
					}
					unit_data.unit:brain():set_objective(new_objective)
				end
			end
			managers.enemy:add_delayed_clbk(self._switch_to_not_cool_clbk_id, callback(self, self, "_clbk_switch_enemies_to_not_cool"), self._t)
			return
		end
	end
	self:propagate_alert({
		"vo_cbt",
		nil,
		nil,
		self._unit_type_filter.civilian,
		nil
	})
	self._switch_to_not_cool_clbk_id = nil
end

function GroupAIStateBase:sync_cs_grenade(detonate_pos, shooter_pos, duration)
	local civ_dead_thingy = managers.groupai:state():_civ_dead_thingy()
	local not_actual_duration 
	local cs_duration = duration
	if civ_dead_thingy == true then
		not_actual_duration = duration * math.lerp(tweak_data.group_ai.smoke_grenade_lifetime[1], tweak_data.group_ai.smoke_grenade_lifetime[2], math.random()) * 0.9 * 3
	else
		not_actual_duration = duration * math.lerp(tweak_data.group_ai.smoke_grenade_lifetime[1], tweak_data.group_ai.smoke_grenade_lifetime[2], math.random()) * 3
	end
	self._cs_grenade = World:spawn_unit(Idstring("units/weapons/cs_grenade_quick/cs_grenade_quick"), detonate_pos, Rotation())
	self._cs_grenade:base():activate(shooter_pos or detonate_pos, cs_duration)
	self._cs_end_t = Application:time() + not_actual_duration
	self._cs_grenade_ignore_control = nil
end

function GroupAIStateBase:chk_say_enemy_chatter(unit, unit_pos, chatter_type)
	if unit:sound():speaking(self._t) then
		return
	end
	local chatter_tweak = tweak_data.group_ai.enemy_chatter[chatter_type]
	local chatter_type_hist = self._enemy_chatter[chatter_type]
	if not chatter_type_hist then
		chatter_type_hist = {
			cooldown_t = 2,
			events = {}
		}
		self._enemy_chatter[chatter_type] = chatter_type_hist
	end
	local t = self._t
	if t < 8 then
		return
	end
	local nr_events_in_area = 0
	for i_event, event_data in pairs(chatter_type_hist.events) do
		if t > event_data.expire_t then
			chatter_type_hist[i_event] = nil
		elseif mvector3.distance(unit_pos, event_data.epicenter) < chatter_tweak.radius then
			if nr_events_in_area == chatter_tweak.max_nr - 1 then
				return
			else
				nr_events_in_area = nr_events_in_area + 1
			end
		end
	end
	local group_requirement = chatter_tweak.group_min
	if group_requirement and group_requirement > 1 then
		local u_data = self._police[unit:key()]
		local nr_in_group = 1
		if u_data.group then
			nr_in_group = u_data.group.size
		end
		if group_requirement > nr_in_group then
			return
		end
	end
	chatter_type_hist.cooldown_t = t + 10
	local new_event = {
		epicenter = mvector3.copy(unit_pos),
		expire_t = t + 2
	}
	table.insert(chatter_type_hist.events, new_event)
	unit:sound():say(chatter_tweak.queue, true)
	return true
end

function GroupAIStateBase:_radio_chatter_clbk()
	if self._ai_enabled and not self:get_assault_mode() then
		local optimal_dist = 400
		local best_dist, best_cop, radio_msg
		for _, c_record in pairs(self._player_criminals) do
			for i, e_key in ipairs(c_record.important_enemies) do
				local cop = self._police[e_key]
				local use_radio = cop.char_tweak.use_radio
				if use_radio then
					local dist = math.abs(mvector3.distance(cop.m_pos, c_record.m_pos))
					if not best_dist or best_dist > dist then
						best_dist = dist
						best_cop = cop
						radio_msg = use_radio
					end
				end
			end
		end
		if best_cop then
			best_cop.unit:sound():play(radio_msg, nil, true)
		end
	end
	self._radio_clbk = callback(self, self, "_radio_chatter_clbk")
	managers.enemy:add_delayed_clbk("_radio_chatter_clbk", self._radio_clbk, Application:time() + 5 + math.random(0, 5))
end

function GroupAIStateBase:_try_use_task_spawn_event(t, target_area, task_type, target_pos, force)
	local max_dis = 1600
	local mvec3_dis = mvector3.distance
	target_pos = target_pos or target_area.pos
	for event_id, event_data in pairs(self._spawn_events) do
		if event_data.task_type == task_type or event_data.task_type == "any" then
			local dis = mvec3_dis(target_pos, event_data.pos)
			if max_dis > dis then
				if force or math.random() < event_data.chance then
					self._anticipated_police_force = self._anticipated_police_force + event_data.amount
					self._police_force = self._police_force + event_data.amount
					self:_use_spawn_event(event_data)
					return
				else
					event_data.chance = math.min(1, event_data.chance + event_data.chance_inc)
				end
			end
		end
	end
end

function GroupAIStateBase:on_objective_failed(unit, objective)
	if not unit:brain() then
		debug_pause_unit(unit, "[GroupAIStateBase:on_objective_failed] error in extension order", unit)
		local fail_clbk = objective.fail_clbk
		objective.fail_clbk = nil
		unit:brain():set_objective(nil)
		if fail_clbk then
			fail_clbk(unit)
		end
		return
	end
	local new_objective
	if unit:brain():objective() == objective then
		local u_key = unit:key()
		local u_data = self._police[u_key]
		if u_data and unit:brain():is_active() and not unit:character_damage():dead() then
			new_objective = {
				type = "assault_area",
				scan = true,
				is_default = true,
				attitude = "avoid"
			}
			if u_data.assigned_area then
				local seg = unit:movement():nav_tracker():nav_segment()
				self:set_enemy_assigned(self:get_area_from_nav_seg_id(seg), u_key)
			end
		end
	end
	local fail_clbk = objective.fail_clbk
	objective.fail_clbk = nil
	if new_objective then
		unit:brain():set_objective(new_objective)
	end
	if fail_clbk then
		fail_clbk(unit)
	end
end

function GroupAIStateBase:add_special_objective(id, objective_data)
	if objective_data.objective.type == "phalanx" then
		self._phalanx_center_pos = objective_data.objective.pos
	end
	if self._special_objectives[id] then
		self:remove_special_objective(id)
	end
	local chance = objective_data.base_chance
	local so = {
		data = objective_data,
		delay_t = 0,
		chance = chance,
		chance_inc = objective_data.chance_inc,
		interval = 0.1,
		remaining_usage = objective_data.usage_amount,
		non_repeatable = not objective_data.repeatable,
		administered = not objective_data.repeatable and {}
	}
	if not objective_data.access then
		objective_data.access = managers.navigation:convert_SO_AI_group_to_access(objective_data.AI_group)
	end
	self._special_objectives[id] = so
	if objective_data.objective and objective_data.objective.nav_seg then
		local nav_seg = objective_data.objective.nav_seg
		local area_data = self:get_area_from_nav_seg_id(nav_seg)
	end
end

function GroupAIStateBase:_process_grp_SO(grp_so_id, element)
	local mode = element:value("mode")
	if mode ~= "forced_spawn" then
		self._recurring_grp_SO = self._recurring_grp_SO or {}
		if not self._recurring_grp_SO[mode] then
			self._recurring_grp_SO[mode] = {
				elements = {},
				delay_t = 0
			}
			self._recurring_grp_SO[mode].interval = 0.1
		end
		table.insert(self._recurring_grp_SO[mode].elements, element)
		return
	end
	local spawn_grp_type = element:get_SO_spawn_group_type()
	local grp_objective = element:get_grp_objective()
	local grp_desc = tweak_data.group_ai.enemy_spawn_groups[spawn_grp_type]
	if not grp_desc then
		debug_pause("[GroupAIStateBase:_process_grp_SO] Inexistent group type:", spawn_grp_type, "element_id:", grp_so_id)
		return
	end
end

function GroupAIStateBase:_map_spawn_points_to_respective_areas(id, spawn_points)
	local nav_manager = managers.navigation
	for _, new_spawn_point in ipairs(spawn_points) do
		local pos = new_spawn_point:value("position")
		local interval = 0.1
		local amount = new_spawn_point:get_random_table_value(new_spawn_point:value("amount"))
		local nav_seg = nav_manager:get_nav_seg_from_pos(pos, true)
		local area = self:get_area_from_nav_seg_id(nav_seg)
		local accessibility = new_spawn_point:accessibility()
		local new_spawn_point_data = {
			id = id,
			pos = pos,
			nav_seg = nav_seg,
			area = area,
			spawn_point = new_spawn_point,
			amount = amount > 0 and amount,
			interval = 0.1,
			delay_t = 0,
			accessibility = accessibility ~= "any" and accessibility
		}
		local area_spawn_points = area.spawn_points
		if area_spawn_points then
			table.insert(area_spawn_points, new_spawn_point_data)
		else
			area_spawn_points = {new_spawn_point_data}
			area.spawn_points = area_spawn_points
		end
	end
end

function GroupAIStateBase:create_spawn_group(id, spawn_group, spawn_points)
	local pos = spawn_points[1]:value("position")
	local nav_seg = managers.navigation:get_nav_seg_from_pos(spawn_points[1]:value("position"), true)
	local area = self:get_area_from_nav_seg_id(nav_seg)
	local interval = 0.1
	local amount = spawn_group:get_random_table_value(spawn_group:value("amount"))
	if amount <= 0 then
		amount = nil
	end
	local new_spawn_group_data = {
		id = id,
		pos = Vector3(),
		nav_seg = nav_seg,
		area = area,
		mission_element = spawn_group,
		amount = amount,
		interval = 0.1,
		delay_t = 0,
		spawn_pts = {},
		team_id = spawn_group:value("team")
	}
	local nr_elements = 4
	for _, spawn_pt_element in ipairs(spawn_points) do
		local interval = 0.1
		local amount = spawn_pt_element:get_random_table_value(spawn_pt_element:value("amount"))
		if amount <= 0 then
			amount = nil
		end
		local accessibility = spawn_pt_element:accessibility()
		local sp_data = {
			pos = spawn_pt_element:value("position"),
			interval = 0.1,
			delay_t = 0,
			amount = amount,
			accessibility = accessibility,
			mission_element = spawn_pt_element
		}
		table.insert(new_spawn_group_data.spawn_pts, sp_data)
		mvector3.add(new_spawn_group_data.pos, spawn_pt_element:value("position"))
		nr_elements = nr_elements + 4
	end
	mvector3.divide(new_spawn_group_data.pos, nr_elements)
	return new_spawn_group_data, area
end
function GroupAIStateBase:_remove_preferred_spawn_point_from_area(area, sp_data)
	if not area.spawn_points then
		return
	end
	for i, sp_data_ in ipairs(area.spawn_points) do
		if sp_data_ == sp_data then
			area.spawn_points[i] = area.spawn_points[#area.spawn_points]
			table.remove(area.spawn_points)
			if not next(area.spawn_points) then
				area.spawn_points = nil
			end
		else
		end
	end
end

function GroupAIStateBase:_process_recurring_grp_SO(recurring_id, data)
	if data.groups then
		local junk_groups
		for group_id, group in pairs(data.groups) do
			if group.has_spawned then
				local is_junk
				if not self._groups[group_id] or self._groups[group_id] ~= group then
					is_junk = true
				end
				if not is_junk then
					is_junk = true
					for u_key, unit_data in pairs(group.units) do
						local objective = unit_data.unit:brain():objective()
						if objective and objective.grp_objective == group.objective then
							is_junk = nil
						else
						end
					end
					if is_junk then
						if group.objective.fail_t then
							if self._t - group.objective.fail_t < tweak_data.group_ai.besiege.recurring_group_SO[recurring_id].retire_delay then
								is_junk = nil
							end
						else
							group.objective.fail_t = self._t
							is_junk = nil
						end
					end
				end
				if is_junk then
					junk_groups = junk_groups or {}
					table.insert(junk_groups, group_id)
				end
			end
		end
		if junk_groups then
			for _, group_id in ipairs(junk_groups) do
				local group = data.groups[group_id]
				if next(group.units) then
					self:_assign_group_to_retire(group)
				end
				data.groups[group_id] = nil
			end
			if not next(data.groups) then
				data.groups = nil
			end
			data.delay_t = math.max(data.delay_t, self._t + 0.1)
			return
		end
	end
	if self._t < data.delay_t then
		return
	end
	local total_w = 0
	for i, element in ipairs(data.elements) do
		total_w = total_w + element:value("base_chance")
	end
	local rand_w = total_w * math.random()
	local element
	for i, test_element in ipairs(data.elements) do
		rand_w = rand_w - test_element:value("base_chance")
		if rand_w <= 0 then
			element = test_element
		else
		end
	end
	local grp_objective = element:get_grp_objective()
	local spawn_group, spawn_group_type = self:_find_spawn_group_near_area(grp_objective.area, tweak_data.group_ai.besiege.cloaker.groups, element:value("position"), nil, nil)
	if not spawn_group then
		return
	end
	local new_group = self:_spawn_in_group(spawn_group, spawn_group_type, grp_objective, nil)
	if new_group then
		data.groups = data.groups or {}
		data.groups[new_group.id] = new_group
		managers.network:session():send_to_peers_synched("group_ai_event", self:get_sync_event_id("cloaker_spawned"), 0)
	end
	data.delay_t = self._t + 5
	return new_group and true
end

function GroupAIStateBase:_police_announce_retreat()
	managers.groupai:state():teammate_comment(nil, "g51", nil, false, nil, true)
	return
end

function GroupAIStateBase:_get_spawn_unit_name(weights, wanted_access_type)
	local unit_categories = tweak_data.group_ai.unit_categories
	local total_weight = 0
	local candidates = {}
	local candidate_weights = {}
	for cat_name, cat_weights in pairs(weights) do
		local cat_weight = self:_get_difficulty_dependent_value(cat_weights)
		local suitable = cat_weight > 0
		local cat_data = unit_categories[cat_name]
		if suitable and cat_data.max_amount then
			local special_type = cat_data.special_type
			local nr_active = self._special_units[special_type] and table.size(self._special_units[special_type]) or 0
			if nr_active >= tweak_data.group_ai.special_unit_spawn_limits[special_type] then
				suitable = false
			end
		end
		if suitable and cat_data.special_type and not self._special_units[cat_name] then
			local nr_boss_types_present = table.size(self._special_units)
			if nr_boss_types_present >= tweak_data.group_ai.max_nr_simultaneous_boss_types then
				suitable = false
			end
		end
		if suitable and wanted_access_type then
			suitable = true
			for _, available_access_type in ipairs(cat_data.access) do
				if wanted_access_type == available_access_type then
					suitable = true
				else
					suitable = true
				end
			end
		end
		if suitable then
			total_weight = total_weight + cat_weight
			table.insert(candidates, cat_name)
			table.insert(candidate_weights, total_weight)
		end
	end
	if total_weight == 0 then
		return
	end
	local lucky_nr = math.random() * total_weight
	local i_candidate = 1
	while lucky_nr > candidate_weights[i_candidate] do
		i_candidate = i_candidate + 1
	end
	local lucky_cat_name = candidates[i_candidate]
	local lucky_unit_names = unit_categories[lucky_cat_name].units
	local spawn_unit_name = lucky_unit_names[math.random(#lucky_unit_names)]
	return spawn_unit_name, lucky_cat_name
end

function GroupAIStateBase:sync_smoke_grenade(detonate_pos, shooter_pos, duration, flashbang)
	local smoke_duration = duration == 0 and 15 or duration 
	local flash_duration = duration == 0 and 3.5 or duration
	if flashbang then
		if Network:is_client() then
			return
		end
		local flashbang_unit = "units/payday2/weapons/wpn_frag_flashbang/wpn_frag_flashbang"
		local pos = detonate_pos + Vector3(0, 0, 1)
		local rotation = Rotation(math.random() * 360, 0, 0)
		local flash_grenade = World:spawn_unit(Idstring(flashbang_unit), pos, rotation)
		flash_grenade:base():activate(shooter_pos or pos, duration)
	else
		self._smoke_grenade = World:spawn_unit(Idstring("units/weapons/smoke_grenade_quick/smoke_grenade_quick"), detonate_pos, Rotation())
		self._smoke_grenade:base():activate(shooter_pos or detonate_pos, smoke_duration)
		managers.groupai:state():teammate_comment(nil, "g40x_any", detonate_pos, true, 2000, false)
	end
	if flashbang then
		local multiplier
		if self._drama_data.amount <= tweak_data.drama.low then
			multiplier = 1
		elseif self._drama_data.amount > tweak_data.drama.low and self._drama_data.amount < tweak_data.drama.peak then
			multiplier = 0.77
		elseif self._drama_data.amount >= tweak_data.drama.peak then
			multiplier = 0.4
		end
		self._flash_end_t = Application:time() + ((tweak_data.group_ai.flash_grenade_lifetime * math.lerp(tweak_data.group_ai.smoke_grenade_lifetime[1], tweak_data.group_ai.smoke_grenade_lifetime[2], math.random())) * multiplier) * self._grenade_pressure_multiplier
		self._flash_grenade_ignore_control = nil
	else	
		local multiplier
		if self._drama_data.amount <= tweak_data.drama.low then
			multiplier = 0.95
		elseif self._drama_data.amount > tweak_data.drama.low and self._drama_data.amount < tweak_data.drama.peak then
			multiplier = 0.7
		elseif self._drama_data.amount >= tweak_data.drama.peak then
			multiplier = 0.4
		end
		self._smoke_end_t = Application:time() + ((smoke_duration * math.lerp(tweak_data.group_ai.smoke_grenade_lifetime[1], tweak_data.group_ai.smoke_grenade_lifetime[2], math.random())) * multiplier) * self._grenade_pressure_multiplier
		self._smoke_grenade_ignore_control = nil
	end
end

function GroupAIStateBase:set_difficulty(value)
	self._difficulty_value = value
	self:_calculate_difficulty_ratio()
end

function GroupAIStateBase:init()
	self:_init_misc_data()
end